from waitress import serve
from api.api import API
from whitenoise import WhiteNoise
import os
def render(app_routes,host="127.0.0.1",port=8001,url_prefix=""):
    # all other app configs to be written here
    curr = os.getcwd()
    static = os.path.join(os.path.dirname(__file__),"router")
    print(f"=======static---->{static}")
    app_ = API(app_routes)
    app_ = WhiteNoise(app_,root=static,prefix="assets/")
    #app_.add_files(static,prefix="/")
    """
    for k,v in app_routes.items():
        if app_routes.routes[k]["static"]:
            # add to whitenoise
            if isinstance(,str):
                
                app_.add_files(app_routes.routes[k]["static"],prefix='') if not app_routes.routes[k]["prefix"] else app_.add_files(app_routes.routes[k]["prefix"],prefix=f'{app_routes.routes[k]["prefix"]}/')
            elif isinstance(,list):
                for static in app_routes.routes[k]["static"]:
                    app_.add_files(app_routes.routes[k]["static"],prefix='') if not app_routes.routes[k]["prefix"] else app_.add_files(app_routes.routes[k]["prefix"],prefix=f'{app_routes.routes[k]["prefix"]}/')
    """
    #print(f"======APP Routes :: {app.routes}")
    try:
        serve(app_,host=host,port=port,url_prefix=url_prefix)
        print("<--------------------Hotweb : Powered by ManlowCharumbira---------------------->")
        print(f"                    APP RUNNING ON --> {host}:{port}")
        print("<--------------------HAPPY BLOGGING-------------------------------------------->")
    except Exception as e:
        print(f"An error occured while trying to serve the app :: {e}")


class Router:
    routes = {"/404":{"404":"4040 page","compo":6}}
    def __init__(self):
        pass
    """
    later on to handle dynamic urls   
    """
    def route(self,path,component,static = None,prefix=None):
        # to be updated for class and functional components
        self.routes[path] = {}
        self.routes[path][path] = component
        self.routes[path]["static"] = static
        self.routes[path]["prefix"] = static
                




# models functions

class Models:
    def __init__(self,model_name,db=None,dialect=""):
        self.model_name =model_name
        self.db = db
    # validate if the user key_queries are accurate
    """
    check if the key value pairs of queries match eg if a field is null but
    its required thru model declaration then raise an error
    ---------------------------------------------------------------------------
    also check if the datatypes are the same
    ---------------------------------------------------------------------------
    """
    def _validate_input(self,arg):
        for col,attributes in self.table_fields.items():
            # attributes is a dictionary
            if "allowNull" in attributes.keys():
                if col["allowNull"] == False and col not in arg.keys():
                    # raise field required error
                    pass

                elif col["allowNull"] == False and col not in arg.keys():
                    # raise field canot be empty error
                    pass
    def _validate_input_type(self,arg):
        pass
    # args is a dict of model fields
    def create(self,args,many={}):
        # many will handle executemany()
        db = self.db
        curr = db.cursor()
        len_args = len(args.keys())
        plcaholders= "("+"%,"*(len_args-1)+"%"+")"
        cols = "("
        for col in args.keys():
            cols =cols + col+","
        values = tuple(args.values())
        print(tuple(values))
        query = f"INSERT INTO {self.model_name}" +cols[:-1] +")" +" VALUES"+ plcaholders
        try:
            curr.execute(query,values)
            db.commit()
        except Exception as e:
            print(e)
        finally:
            db.close()
        print(query)
    # used to unpack the args it returns the col names
    def _unpack_args(self,args):
        pass
    # sketch of the query
    def findOne(self,args):
        if not self.db:
            # raise no db connection estabilished error
            pass
        where_placeholders = ""
        values = []
        for k,v in args.items():
            if k.lower() == "where":
                for col,val in args[k].items():
                    where_placeholders = where_placeholders+f" {col}=%s AND "
                    values.append(val)
        arg_cols = "*"
        curr = self.db.cursor()
        place_holders = ""
        vals = ""
        query = f"SELECT {arg_cols} FROM {self.model_name}"+" WHERE"+place_holders 
        q_exec = curr.execute(query,vals)
        results = q_exec.fetchOne()
        self.db.close()
        return results
    
    def findAll(self,arg):
        pass
    # to delete
    def destroy(self,arg):
        pass
    
    def findAndUpdate(self,args):
        pass
    # used to create fields for the db, not used to create tables created thru migrations
    def init(self,model_name,args):
        if not isinstance(args,dict):
            # raise an error expecting a dict
            pass

        query = "create table query"
        self.model_name=model_name
        self.table_fields =args

        # table_fields = ""
        # len_args = len(args)
        # sql = ""
    
    def _gettype_(self,arg):
        if len(arg)==0:
            return False
        if isinstance(arg,str):
            return "%s"
        elif isinstance(arg,int):
            return "%d"
    """
    hasOne and has Many are used to define associations
    model arg --> is a string name for a Model
    """
    def hasOne(self,model,options={}):
        pass
    
    def hasMany(self,model,options={}):
        pass

class DBConnection:
    """
    dialect defines the type of sql database
    args define a dictionary of db configurations eg host,port,password
    try to connect to db else raise an error
    -->install the driver for the selected dialect except for sqlite which is built in
    """
    
    def __init__(self,dialect,**kwargs):
        self.dialect = dialect
        self.kwargs = kwargs
    def _unpack_kwargs(self):
        pass
    
    def connect(self):
        # call appropriate method based on provided dialect
        if self.dialect.lower() =="sqlite" or self.dialect.lower() =="sqlite3":
            self.sqlite_db()
        elif self.dialect.lower() =="mysql":
            self.mysql_db()
        pass
    def sqlite_db(self):
        import sqlite3
        # get db name from the kwargs

        conn = sqlite3.connect(self.kwargs)
        return conn
    
    def mysql_db(self):
        import mysql.connector as mysql_
        args={}
        db = mysql_.connect(
            host=args["localhost"],
            user=args["username"],
            password=args["password"],
            database=args["database"]
            )
        return db
    def postgress_db(self):
        pass
    def mssql_db(self):
        pass
    
    # other dbs


class Models_Migrate:
    pass

"""
------from test.py not here-------to be deleted----------
table={
    "username":{
        "type":"STRING",
        "allowNull":"False",
        "length":"12"
    },
    "password":{
        "type":"text",
    }
}
def col_func(args,col,attr):
    line = f"{col} "
    for key in args[col].keys():
        if key == "length":
            continue
        if key =="type":
            if "length" in args[col].keys():
                type_ = f"{args[col]['type']}({args[col]['length']})" 
                line = line +" "+ type_ +" "
            else:
                length = "15"
                type_ = f"{args[col]['type']}({length})" 
                line = line +" "+ type_ +" "
            continue
        if key.lower() == "allownull":
            if args[col][key]=="False":
                null = "NOT NULL"
                line = line +" "+ null
            continue
        line = line +" "+ key +" "
    return line
def unpack_table(args,t_name="users"):
    table_ = F"CREATE TABLE {t_name}("
    end_table = ")"
    line = ""
    for col,attr in args.items():
        """for key in args[col].keys():
            line = line+ " "+ " "+ key + args[col][key] + """
        line = line+ col_func(args,col,attr)+", \n"
        print(col,attr)
        #table_ += col_func(col[attr],attr)
    print(table_ + line +end_table )
    return table_
#print(unpack_table(table))
ins={
    "username":"chaxes",
    "pana":"manlow",
    "user_id":"12"
}
def insert(args):
    len_args = len(args.keys())
    plcaholders= "("+"%,"*(len_args-1)+"%"+")"
    cols = "("
    for col in args.keys():
        cols =cols + col+","
    values = args.values()
    print(tuple(values))
    query = "INSERT INTO table_name" +cols[:-1] +")" +" VALUES"+ plcaholders
    print(query)
def select(args):
    where_placeholders = ""
    values = []
    cols ="*"
    
    for k,v in args.items():
        if k.lower() =="attributes":
            cols = ""
            cols = cols + " ,".join(args[k])
        if "attributes" not in args.keys():
            cols = "*"
        if k.lower() == "where":
            for col,val in args[k].items():
                where_placeholders = where_placeholders+f" {col}=%s AND "
                values.append(val)
    query = f"SELECT {cols} FROM table_name"+" WHERE "
    full_select = query +where_placeholders[:-4]
    print(full_select,tuple(values))
def update(args):
    pass
def delete(args):
    pass
selec ={
    "where":{
        "user":"panashe",
        "pass":"1234"
    },
    "attributes":["col1","col..."],
    "where2":{
        "Op.Or":{
            "col1":"val1..."
        }
    }
}
select(selec)
#insert(ins)
va = ["hello","manlow"]
val = "------>".join(va)
print(val)
"""




class Footer:
    def __init__(self):
        pass
    
    def footer(self,content="Brought  To You With Love from HotWeb",files=[]):
        open_footer_tag = "\n" + "<footer>" + "\n"
        close_footer_tag = "\n" + "</footer>" + "\n"
        close_body_tag = "\n" + "</body>" + "\n"
        concat = "\n "
        if len(files)>0:
            for file in files:
                concat = concat + "\n" + file
        open_footer_tag = open_footer_tag + content + close_footer_tag + close_body_tag + concat

        return open_footer_tag


class HeadContent:
    def __init__(self,title="Document",files=[]):
        self.files = files
        self.head(title)
    def head(self,title):
        concat = "\n"
        end_head = "\n" + "</head>"
        if len(self.files) >0:
            for file in self.files:
                concat = concat + "\n" + file
        self.part_one = f"""
                        <!DOCTYPE html>
                        <html lang="en">
                        <head>
                        <meta charset="UTF-8">
                        <meta name="viewport" content="width=device-width, initial-scale=1.0">
                        <meta http-equiv="X-UA-Compatible" content="ie=edge">
                        <title>{title}</title> 
                         """
        self.part_one = self.part_one + concat + end_head
        return self.part_one
hd = HeadContent(files=["css","bootstrap"])


"""
to handle whitenoise static files serving
"""
import os
import gzip
#os.path.dirname(__file__)
INPUT_PATH = os.path.join(os.getcwd(),"static")
OUTPUT_PATH = os.path.join(os.getcwd(),"staticfiles")
SKIP_COMPRESS_EXTENTIONS =[
    # FILE EXT
    #images
    ".jpg",
    ".jpeg",
    ".png",
    ".gif",
    ".webp",
    #compressed files
    ".zip",
    ".gz",
    ".tgz",
    ".bz2",
    ".tbz",
    ".xz",
    ".br",
    #Flash
    ".swf",
    ".flv",
    # fonts
    ".woff",
    ".woff2",
]
def remove_files(path):
    print(f"removing files from {path}")
    for filename in os.listdir(path):
        file_path = os.path.join(path,filename)
        try:
            if os.path.isfile(file_path):
                os.unlink(file_path)
        except Exception as e:
            print(e)

def main():
    # remove all files from the staticfiles dir
    if not os.path.exists(OUTPUT_PATH):
        os.makedirs(OUTPUT_PATH)
    remove_files(OUTPUT_PATH)
    
    for dirpath,dirs,files in os.walk(INPUT_PATH):
        for filename in files:
            input_file = os.path.join(dirpath,filename)
            with open(input_file,"rb") as f:
                data = f.read()
            # compress if file extenstion not in SKIP COMPRESS list
            name,ext = os.path.splitext(filename)
            if ext not in SKIP_COMPRESS_EXTENTIONS:
                # save compressed file to static files dir folder
                compressed_output_file = os.path.join(OUTPUT_PATH,f"{filename}.gz")
                print(f"Compressing {filename} ... ")
                print(f"Saving {filename}.gz ")
                output = gzip.open(compressed_output_file,"wb")
                try:
                    output.write(data)
                finally:
                    output.close()
            else:
                print(f"skipping compressing {filename} ")
                # save original file to the staticfiles folder
                output_file = os.path.join(OUTPUT_PATH,filename)
                with open(output_file,"wb") as f:
                    f.write(data)

main()


"""
this class is used to create dom function eg.

function sample_dom_function(){
    // bla bla bla
    // dom manipulation
}
"""
class DomFunction:
    def __init__(self,name="sample_dom_function"):
        pass
    def add_child(self,child):
        pass


class Dom:
    js_template = ""
    def __init__(self):
        pass
    
    def click(self,class_id,fn):
        pass
    def write_to_js(self,js=""):
        with open("dom.js","w") as f:
            file = self.generate_js_file(js)
            f.write(file)
    def generate_js_file(self,js):
        start = """
        $(document).ready(function(){
        """
        end = "\n \t" + "})"
        if len(js)>0:
            end = js + "\n" + end
        content = start + end
        return content
    def removeClass(self,klass):
        pass

d = Dom()
#d.write_to_js("// d instance call")
#e = Dom()
#e.write_to_js("// e instance call")
import inspect
clk = str(inspect.getsource(d.click))
print(clk + "\t hello")

"""
/* 
         the json file keeps track of two values
         1. commited state=> this means content was written to js
         2. to_commited state
*/
"""


from server import render
from shortcuts.load_file import current_dir
from router.route_ import Router
print(current_dir)
import os
output = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta http-equiv="X-UA-Compatible" content="ie=edge">
    <title>Document</title>
    <link rel="stylesheet" href="assets/test.css">
</head>
<body>
  <h1>Manlow</h1>  
  <img src='/assets/section.PNG' alt='hello' />
</body>
</html>
"""
def eg(req):
    return "hello from manlow"
app = Router()
def hel(req):
    img = os.path.join(os.getcwd(),"router","section.png")
    
    print(os.getcwd())
    return output #"<h1>Hello There</h1>" + "<br>" + f"<img src='/assets/section.PNG' alt='hello' />"
app.route("/",hel)
app.route("/hello",eg)
render(app)


import os

current_dir = os.getcwd()
#print(f"=={current_dir}")
def load(path):
    new_dir = os.chdir(path)
    #print(new_dir)
    file = os.path.join(str(os.getcwd()),"views","all.py")
    print(file)
    print("The new path====",os.getcwd())
#print(dir(os))
#load("../engine/views")
#print(f"=={current_dir}")

#print(dir(os.path.dirname))
print(os.path.dirname(os.path.join(str(os.getcwd()),"views","all.py")).split(os.path.sep))


from setuptools import setup, find_packages
desc = """
## HOTWEB WEB FRAMEWORK
Python web framework for designing lucrative websites in python programming language. The goal of HOTWEB FRAMEWORK 
is to allow developers to develope websites in one language, that is python because you can do anything using Python Programming language

## Author
```{py}
Real Manlow (aka ManlowCharumbira)
```
## Contacts
instagram.com/realmanlow
realmanlow20@gmail.com
charumbiramanlow20@gmail.com
"""
def readme():
    with open("readme.md","r") as f:
        info = f.read()
        return info
setup(
    name="hotweb",
    version="0.7.182",
    license="MIT",
    author = "Real Manlow,aka ManlowCharumbira",
    long_description=desc,
    long_description_content_type="text/markdown",
    author_email="realmanlow20@gmail.com",
    packages=find_packages(),
    #package_dir = {'':'hotweb'},
    include_package_data = True,
    keywords = "hotweb python-web-framework python web framework fast light secure",
    install_requires = [
        "parse","waitress","webob","requests","hotweb-cli","whitenoise","Werkzeug","Jinja2","hotweb-cli"
    ]
)


class Form:
    
    def __init__(self,component="form",**kwargs):
        self.kwargs = kwargs
        self.component =component
        self.has_closing = True
        self.current_component = f"< {self.component} "
        self.current_end ='\n' + f"</{self.component}>"
        #self.current_component = current_component + self.current_end
    def __call__(self,component):
        print("============inside call")
        self.define_component(component)
        self.build_component()
    # append child element to the current component
    def add_child(self,child):
        # slice from the closing tag to append the child before closing tag
        end_len = len(self.current_end)
        self.current_end =  '\n' + child + self.current_end
    # remove child element to the current component,can get help from javascript
    def remove_child(self,child):
        pass
    def add_class(self):
        pass
    def remove_class(self):
        pass
    # declare type of component to be built
    def define_component(self,component,has_closing=True):
        # to be updated to validate types of components entered
        self.has_closing = has_closing
        self.component = component
        return self.component
    # now building the component and its attributes
    def build_component(self):
        # content of the html component
        content = ""
        has_class = False
        attributes = ''
        for attr,value in self.kwargs.items():
            if attr == "content":
                content = content + attr
                continue
            if attr == "class_":
                attributes += f'class="{self.component}__class component_class {value}" '
                has_class = True
                continue
            attributes = attributes + ' ' + f'{attr}="{value}" '
        if not has_class:
            attributes += f'class="{self.component}__class" '
        if self.has_closing:
             self.current_component = self.current_component + attributes + " >" + "\n" + content + "\n" + self.current_end
             return self.current_component
        else:
            self.current_component = self.current_component + attributes + " />" + "\n"
            return self.current_component

def Button():
    return "<div>INSIDE A DIV ELEMENT</div>"
form = Form(method="post",action="localhost/posts",class_="hello test class")
form.add_child(Button())
form.add_child("<div>INSIDE A DIV ELEMENT</div> \n")
print(form.build_component())
text = "</manlow"
#text[2] = text[2] + "b"
#t1 = text[1:2]
#print(text[-1:].split(" </"))
print()


"""
to define db queries for different sql dialect
"""
from models import DBConnection
# to get db connection
# read the db configs from the command line
class QueryInterface:
    def __init__(self,db_instance):
        self.db = db_instance
    
    def createTable(self,table_name,attrs):
        pass
    
    def dropTable(self,table_name):
        pass
    def addColumn(self,col):
        pass
    def dropColumn(self,col):
        pass
    def alterColumn(self,col):
        pass
    def addIndex(self,col):
        pass

queryInterface = QueryInterface("test_db_args")